import re
import requests
from collections import namedtuple
from urllib.parse import ParseResult, urlencode, urlparse
from .oauth_token import OAuthToken

__all__ = ['NexaasIDOAuthClient']


BaseOAuthClient = namedtuple(
    'NexaasIDOAuthClient',
    'id secret scope redirect_uri server',
)


class NexaasIDOAuthClient(BaseOAuthClient):

    def __new__(cls, client_id: str, secret: str, *,
                scope: str = None, redirect_uri: str,
                server: str = None) -> BaseOAuthClient:
        server = server or 'http://localhost:3000/'
        scope = scope or 'profile'
        if not re.match(r'^[a-z]+://', server):
            server = 'https://' + server

        return super().__new__(cls, client_id, secret, scope, redirect_uri,
                               urlparse(server))

    @property
    def authorize_url(self) -> str:
        query = urlencode({
            'response_type': 'code',
            'client_id': self.id,
            'redirect_uri': self.redirect_uri,
            'scope': self.scope,
        })
        return self.server._replace(
            path='/oauth/authorize',
            query=query,
        ).geturl()

    def __get_oauth_token(self, **kwargs) -> OAuthToken:
        post_data = {
            'client_id': self.id,
            'client_secret': self.secret,
            'redirect_uri': self.redirect_uri,
        }
        post_data.update(kwargs)
        res = requests.post(
            self.server._replace(path='/oauth/token').geturl(),
            post_data,
        )
        res.raise_for_status()
        try:
            return OAuthToken(**res.json())

        except (ValueError, KeyError) as exc:
            new_exc = ValueError('no access token supplied')
            new_exc.__context__ = exc
            raise new_exc

    def get_token(self, code: str = None) -> OAuthToken:
        if code:
            return self.__get_oauth_token(
                grant_type='authorization_code',
                code=code,
            )
        else:
            return self.__get_oauth_token(grant_type='client_credentials')

    def refresh_token(self, token: OAuthToken) -> OAuthToken:
        return self.__get_oauth_token(
            grant_type='refresh_token',
            refresh_token=token.refresh_token,
        )



from django.conf import settings
from django.shortcuts import redirect
from django.urls import reverse
from django.views.decorators.cache import never_cache
from nexaas_id_client import NexaasIDOAuthClient
from nexaas_id_client.oauth_token import TokenSerializer

__all__ = ['signin', 'callback']


def signin(request):
    next_url = request.GET.get('next_url') or \
               request.META.get('HTTP_REFERER')
    if next_url:
        request.session['oauth_next_url'] = next_url
    return redirect(get_client(request).authorize_url)


def signout(request):
    session = request.session
    if 'oauth_token' in session:
        del session['oauth_token']
    next_url = request.GET.get('next_url') or \
               session.get('oauth_next_url') or \
               request.META.get('HTTP_REFERER') or \
               '/'
    return redirect(next_url)


def callback(request):
    client = get_client(request)
    code = request.GET.get('code')
    session = request.session
    session['oauth_token'] = TokenSerializer.serialize(client.get_token(code))
    next_url = session.get('oauth_next_url')
    if next_url:
        del session['oauth_next_url']
    return redirect(next_url or '/')


def get_client(request):
    return NexaasIDOAuthClient(
        settings.NEXAAS_ID_CLIENT_ID,
        settings.NEXAAS_ID_CLIENT_SECRET,
        server=settings.NEXAAS_ID_HOST,
        redirect_uri=request.build_absolute_uri(reverse(callback)),
        scope=settings.NEXAAS_ID_CLIENT_SCOPE,
    )



from abc import ABCMeta
from collections import namedtuple
from datetime import datetime, timedelta
from urllib.parse import parse_qsl, urlencode

__all__ = ['OAuthToken', 'TokenSerializer']

# XXX: for test purpose
# TODO: remove this when updating to python 3.5+
_isinstance = isinstance


class OAuthToken(metaclass=ABCMeta):

    Base = namedtuple(
        'OAuthToken',
        'access_token refresh_token expires_at scope',
    )

    def __new__(_cls, *args, **kwargs) -> Base:
        return MainOAuthToken(*args, **kwargs)

    @classmethod
    def __subclasshook__(cls, C):
        for attr in cls.Base._fields:
            if not any(attr in B.__dict__ for B in C.__mro__):
                return NotImplemented
        return True


class MainOAuthToken(OAuthToken.Base):

    def __new__(cls, access_token: str, refresh_token: str = None,
                expires_at: datetime = None,
                created_at: int = -1, expires_in: int = -1,
                scope: str = 'profile', **__) -> OAuthToken.Base:
        if not expires_at:
            if expires_in >= 0:
                if created_at >= 0:
                    created_at = datetime.fromtimestamp(created_at)
                else:
                    created_at = datetime.now()
                expires_at = created_at + timedelta(seconds=expires_in)

            else:
                expires_at = None
        return super().__new__(
            cls, access_token, refresh_token, expires_at, scope,
        )

    @property
    def expired(self) -> bool:
        return self.expires_at < datetime.now()

OAuthToken.register(MainOAuthToken)


class TokenSerializer:

    @staticmethod
    def serialize(token: OAuthToken) -> str:
        return urlencode({
            attr: getattr(token, attr) or ''
            for attr in OAuthToken.Base._fields
        })

    @staticmethod
    def deserialize(token: str) -> OAuthToken:
        resource = {
            attr: value or None
            for attr, value in parse_qsl(token)
            if attr in OAuthToken.Base._fields
        }

        if resource.get('expires_at'):
            resource['expires_at'] = datetime.strptime(
                resource['expires_at'],
                r'%Y-%m-%d %H:%M:%S',
            )
        return OAuthToken(**resource)



from django.urls import path
from . import views

__all__ = ['urlpatterns']

urlpatterns = [
    path(r'signin/', views.signin, name='nexaas-id-signin'),
    path(r'signout/', views.signout, name='nexaas-id-signout'),
    path(r'callback/', views.callback, name='nexaas-id-callback'),
]



from requests.exceptions import HTTPError
from requests.models import Response

__all__ = ['SignedOutException']


class SignedOutException(HTTPError):

    @classmethod
    def from_http_error(cls, exc: HTTPError) -> HTTPError:
        error = cls(
            request=exc.request,
            response=exc.response,
        )
        error.__context__ = exc
        return error

    @classmethod
    def check_status(cls, res: Response) -> None:
        try:
            res.raise_for_status()
        except HTTPError as exc:
            if res.status_code == 401:
                raise cls.from_http_error(exc)
            raise



from functools import wraps
import inspect
from django.shortcuts import redirect
from django.urls import reverse
from nexaas_id_client import NexaasIDClient
from nexaas_id_client.oauth_token import TokenSerializer
from nexaas_id_client.support.django.helpers import signed_in
from . import views

__all__ = ['authorization_required']


def authorization_required(wrapped):
    """
    XXX: [side effect] Redirect to sign out if the view returns falsy
    FIXME: change this behavior
    """
    @wraps(wrapped)
    def wrapper(request, *args, **kwargs):
        session = get_session(request)
        if not session:
            return redirect(reverse(views.signin))
        kwargs['api_client'] = get_api_client(request, session)
        return deal_with_request(wrapped, request, args, kwargs)
    return wrapper


def get_session(request):
    session = request.session
    return session if 'oauth_token' in session else None


def get_api_client(request, session):
    return NexaasIDClient.from_oauth(
        TokenSerializer.deserialize(session['oauth_token']),
        client=views.get_client(request),
    )


def deal_with_request(wrapped, request, args, kwargs):
    response = None
    with signed_in() as unsigned:
        response = wrapped(request, *args, **kwargs)
    return response or unsigned.redirect



#!/usr/bin/env python

from setuptools import setup, find_packages


setup(
    name = 'nexaas_id_client',
    version = '1.0',
    author = 'Rodrigo Cacilhas',
    author_email = 'rodrigo.cacilhas@nexaas.com',
    description = '',
    license = 'Proprietary License',
    keywords = 'nexaas-id oauth',
    url = '',
    packages = find_packages(exclude=('tests', 'tests.*')),
    long_description = '',
    test_suite = 'tests',
    install_requires = [
        'python-dateutil>=2.7.0',
        'requests>=2.19.0',
    ],
    tests_require=[
        'django==2.0.0',
        'Flask==1.0.2',
        'pycodestyle==2.4.0',
        'pylint==2.0.1',
        'vcrpy==1.13.0',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'License :: Other/Proprietary License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Utilities',
    ],
)



from collections import namedtuple
import re
from typing import Union
from urllib.parse import ParseResult, urlencode, urlparse
from uuid import UUID
import dateutil.parser
import requests
from requests.auth import AuthBase
from requests.models import Request
import warnings
from .oauth_client import NexaasIDOAuthClient
from .oauth_token import OAuthToken
from .exceptions import SignedOutException

__all__ = ['NexaasIDClient']

try:
    DeprecationWarning = warnings.DeprecationWarning
except AttributeError:
    DeprecationWarning = warnings._getcategory('DeprecationWarning')


ClientProps = namedtuple('ClientProps', 'token id secret server')


class NexaasIDClient:
    __slots__ = (
        '__internal_tuple',
        '_personal_info',
        '_professional_info',
        '_emails',
        '_contacts',
    )

    @classmethod
    def from_oauth(cls, token: OAuthToken, *,
                   client: NexaasIDOAuthClient) -> 'NexaasIDClient':
        return cls(
            token=token,
            id=client.id,
            secret=client.secret,
            server=client.server,
        )

    def __init__(self, token: Union[str, OAuthToken], *,
                 id: str = None, secret: str = None,
                 server: Union[str, ParseResult] = None):
        server = server if isinstance(server, ParseResult) \
            else urlparse(server or 'http://localhost:3000/')
        token = token if isinstance(token, OAuthToken) \
            else OAuthToken(access_token=token)
        self.__internal_tuple = ClientProps(token, id, secret, server)
        self.reset()

    def __get_response(self, path: str) -> dict:
        res = requests.get(
            self.server._replace(path=path).geturl(),
            auth=HTTPBearerAuth(self.access_token),
        )
        SignedOutException.check_status(res)
        return res.json(object_hook=_json_decode)

    def __post_response(self, path: str, data: dict) -> dict:
        res = requests.post(
            self.server._replace(path=path).geturl(),
            json=data,
            auth=HTTPBearerAuth(self.access_token),
        )
        SignedOutException.check_status(res)
        return res.json(object_hook=_json_decode)

    @property
    def access_token(self) -> str:
        return self.token.access_token

    @property
    def refresh_token(self) -> str:
        return self.token.refresh_token

    @property
    def scope(self) -> str:
        return self.token.scope

    @property
    def token(self) -> str:
        return self.__internal_tuple.token

    @property
    def id(self) -> str:
        return self.__internal_tuple.id

    @property
    def secret(self) -> str:
        return self.__internal_tuple.secret

    @property
    def server(self) -> ParseResult:
        return self.__internal_tuple.server

    @property
    def personal_info(self) -> 'PersonalInfo':
        if self._personal_info is None:
            info = self.__get_response(path='/api/v1/profile')
            self._personal_info = _build_tuple('PersonalInfo', info)
        return self._personal_info

    @property
    def professional_info(self) -> 'ProfessionalInfo':
        if self._professional_info is None:
            info = self.__get_response(
                path='/api/v1/profile/professional_info'
            )
            self._professional_info = _build_tuple('ProfessionalInfo', info)
        return self._professional_info

    @property
    def emails(self) -> 'Emails':
        if self._emails is None:
            info = self.__get_response(path='/api/v1/profile/emails')
            self._emails = _build_tuple('Emails', info)
        return self._emails

    @property
    def contacts(self) -> 'Contacts':
        if self._contacts is None:
            info = self.__get_response(path='/api/v1/profile/contacts')
            self._contacts = _build_tuple('Contacts', info)
        return self._contacts

    @property
    def navbar_url(self) -> str:
        warnings.warn(
            'deprecated navbar, use user_widget_url',
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.server._replace(
            path='/api/v1/widgets/navbar.js',
            query=urlencode({'access_token': self.access_token}),
        ).geturl()

    @property
    def user_widget_url(self) -> str:
        return self.server._replace(
            path='/api/v1/widgets/user.js',
            query=urlencode({'access_token': self.access_token}),
        ).geturl()

    def invite(self, email: str) -> 'Invitation':
        info = self.__post_response('/api/v1/sign_up', {'invited': email})
        return _build_tuple('Invitation', info)


    def reset(self) -> None:
        self._personal_info = None
        self._professional_info = None
        self._emails = None
        self._contacts = None


class HTTPBearerAuth(AuthBase):

    def __init__(self, token: str):
        self.token = token

    def __eq__(self, other: 'HTTPBearerAuth') -> bool:
        return self.token == other.token

    def __call__(self, r: Request) -> Request:
        r.headers['Authorization'] = 'Bearer {}'.format(self.token)
        return r


def _build_tuple(class_name: str, info: dict) -> tuple:
    info_class = namedtuple(class_name, info.keys())
    return info_class(**info)


def _json_decode(value):
    if isinstance(value, str):
        if _is_date(value):
            return dateutil.parser.parse(value).date()
        if _is_datetime(value):
            return dateutil.parser.parse(value)
        if _is_uuid(value):
            return UUID(value)

    elif isinstance(value, list):
        return [_json_decode(e) for e in value]

    elif isinstance(value, dict):
        return {k: _json_decode(v) for k, v in value.items()}

    return value


def _is_date(value: str) -> bool:
    return re.match(r'^\d{4}-\d\d-\d\d$', value) is not None


def _is_datetime(value: str) -> bool:
    return re.match(
        r'^\d{4}-\d\d-\d\d'
        r'[T ]\d\d:\d\d:\d\d(\.\d{3})?'
        r'([A-Z]{1,5}|GMT[+-]\d+)?$',
        value,
    ) is not None


def _is_uuid(value: str) -> bool:
    return re.match(
        r'^[0-9a-f]{8}-'
        r'[0-9a-f]{4}-'
        r'[0-9a-f]{4}-'
        r'[0-9a-f]{4}-'
        r'[0-9a-f]{12}$',
        value,
    ) is not None



from functools import wraps
import inspect
from flask import Blueprint, current_app, redirect, request, session, url_for
from nexaas_id_client import NexaasIDClient, NexaasIDOAuthClient
from nexaas_id_client.oauth_token import OAuthToken, TokenSerializer
from nexaas_id_client.exceptions import SignedOutException

__all__ = ['authorization_required', 'oauth']


oauth = Blueprint('nexaas_id_oauth', __name__)


def authorization_required(wrapped):
    @wraps(wrapped)
    def wrapper(*args, **kwargs):
        if 'oauth_token' not in session:
            return redirect(url_for('nexaas_id_oauth.signin'))

        client = get_client()
        token = kwargs['api_client'] = NexaasIDClient.from_oauth(
            TokenSerializer.deserialize(session['oauth_token']),
            client=client,
        )

        try:
            return wrapped(*args, **kwargs)

        except SignedOutException:
            try:
                token = client.refresh_token(token)
                session['oauth_token'] = TokenSerializer.serialize(token)
                kwargs['api_client'] = NexaasIDClient.from_oauth(
                    token,
                    client=client,
                )
                return wrapped(*args, **kwargs)

            except SignedOutException:
                return redirect(url_for('nexaas_id_oauth.signout'))
    return wrapper


@oauth.route('/signin')
def signin():
    next_url = request.args.get('next_url') or \
               request.headers.get('Referer')
    if next_url:
        session['oauth_next_url'] = next_url
    return redirect(get_client().authorize_url)


@oauth.route('/signout')
def signout():
    if 'oauth_token' in session:
        del session['oauth_token']
    next_url = request.args.get('next_url') or \
               session.get('oauth_next_url') or \
               request.headers.get('Referer') or \
               '/'
    return redirect(next_url)


@oauth.route('/callback')
def callback():
    client = get_client()
    code = request.args.get('code')
    session['oauth_token'] = TokenSerializer.serialize(client.get_token(code))
    next_url = session.get('oauth_next_url')
    if next_url:
        del session['oauth_next_url']
    return redirect(next_url or '/')


def get_client():
    return NexaasIDOAuthClient(
        current_app.config['NEXAAS_ID_CLIENT_ID'],
        current_app.config['NEXAAS_ID_CLIENT_SECRET'],
        server=current_app.config.get('NEXAAS_ID_HOST'),
        redirect_uri=url_for('nexaas_id_oauth.callback', _external=True),
        scope=current_app.config.get('NEXAAS_ID_CLIENT_SCOPE'),
    )



from contextlib import contextmanager
from django.shortcuts import redirect, reverse
from nexaas_id_client.exceptions import SignedOutException


class signed_in:
    __slots__ = ('redirect')

    def __init__(self):
        self.redirect = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, traceback):
        if exc and exc_type is SignedOutException:
            self.redirect = redirect(reverse('nexaas-id-signout'))
            return True  # inhibit raising expected exception
        return exc
